// BUMP, the imageboard porter
// ===========================
// -This software is designed to help out anon with imageboards

// -Filename: muh_curl.h
// -This is the crude hackery for making libcurl simple to use from BUMP

#pragma once

#include <chrono>
#include <cmath>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <locale>
#include <ratio>
#include <sstream>
#include <string>

using std::cerr;
using std::cout;
using std::fixed;
using std::flush;
using std::locale;
using std::ostringstream;
using std::ratio;
using std::round;
using std::setfill;
using std::setprecision;
using std::setw;
using std::string;
using std::stringstream;
using std::chrono::duration;
using std::chrono::duration_cast;
using std::chrono::hours;
using std::chrono::minutes;
using std::chrono::seconds;
using std::chrono::system_clock;
using std::chrono::time_point;

//------------------------------------------------------------------------------

#include "curl_easy.h"
#include "curl_exception.h"
#include "curl_form.h"
#include "curl_ios.h"

namespace muh_curl {

constexpr int tot_ticks{42};  // how wide the progress meter is (in chars)

/** Our type alias for a @c void*
 */
using Address = void*;

/** Our type alias for a standard @c system_clock @c time_point
 */
using Time = time_point<system_clock>;
// alias related
using Durn = duration<long, ratio<1, 1000000000>>;
// type alias related:
using Call_clk = system_clock;

Time start;  // will start the instant before a DL begins

/** Our type alias for a chrono duration in seconds, cast as float
 */
using Secs_f = duration<float, seconds::period>;

int prt_prog(Address, curl_off_t, curl_off_t, curl_off_t, curl_off_t);

//------------------------------------------------------------------------------

using curl::curl_easy;
using curl::curl_ios;
using curl::curlcpp_traceback;
using curl_excptn = curl::curl_easy_exception;

//------------------------------------------------------------------------------

string proto{"http"};
string res_dir{"res"};
string media_dir{""};

//------------------------------------------------------------------------------
/** Assemble the complete URL from the supplied parts
 * @param site_nm
 * @param board_nm
 * @param page_nm
 * @param pg_type
 * @return
 */
auto muh_url(const string& site_nm,
             const string& board_nm,
             const string& page_nm,
             const string& pg_type) {
  if (page_nm == "catalog")
    return proto + "://" + site_nm + "/" + board_nm + "/" + page_nm + "." +
           pg_type;
  else
    return proto + "://" + site_nm + "/" + board_nm + "/" + res_dir + "/" +
           page_nm + "." + pg_type;
}

//------------------------------------------------------------------------------
/** Assemble the complete URL from the supplied parts
 * @param site_nm
 * @param media_file
 * @return
 */
auto muh_file_url(const string& site_nm, const string& media_file) {
  if (media_dir != "")
    return proto + "://" + site_nm + "/" + media_dir + "/" + media_file;
  else
    return proto + "://" + site_nm + media_file;
}

//------------------------------------------------------------------------------
/** Return the json, or html, or file response to the caller
 * @param url
 * @return
 */
stringstream get_resp(const string& url) {
  stringstream           res{};
  curl_ios<stringstream> writer{res};
  curl_easy              conn{writer};

  conn.add<CURLOPT_URL>(url.data());
  conn.add<CURLOPT_FOLLOWLOCATION>(1L);

  conn.add<CURLOPT_TIMEOUT>(28800L);
  conn.add<CURLOPT_USERAGENT>("bump_v0.2fg");

  // we'll output our DL progress to the console
  conn.add<CURLOPT_XFERINFOFUNCTION>(prt_prog);

  // yes, do go ahead and track our DL progress
  conn.add<CURLOPT_NOPROGRESS>(0L);

  try {  //

    start = Call_clk::now();  // start the timer,
    conn.perform();           // and let's go!

    cout << '\n';

  } catch (const curl_excptn& err) {
    const auto errors{err.get_traceback()};
    std::cerr << '\n'
              << errors[0].first << " : '" << url
              << "'  curl::" << errors[0].second << "()\n";

    // To prevent incomplete or corrupt response streams for files from being
    // written out to disk by the caller.
    //
    if (errors[0].first == "Timeout was reached")
      res.str("9001 Client Timeout");
    else if (errors[0].first == "Transferred a partial file")
      res.str("9002 Partial File");
    else if (errors[0].first == "Stream error in the HTTP/2 framing layer")
      res.str("9003 Stream error");
    else if (errors[0].first == "Couldn't connect to server")
      res.str("9004 Couldn't connect");
  }

  return res;
}

//------------------------------------------------------------------------------
/**
 * @param xfer_frac
 * @return
 */
std::string pct(float const xfer_frac) {
  auto          pct{xfer_frac * 100};
  ostringstream pctss;
  pctss << setw(6) << setfill(' ') << fixed << setprecision(1) << pct;

  return pctss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_frac
 * @param reverse
 * @return
 */
std::string ticks(float const xfer_frac, bool const reverse = false) {
  // part already done (the tickmarks)
  auto done = static_cast<int>(round(tot_ticks * xfer_frac));

  // part still remaining (the spaces)
  auto rem{tot_ticks - done};

  ostringstream tickss;
  if (reverse)
    tickss << setw(rem) << setfill(' ') << "" << setw(done) << setfill('=')
           << "";
  else
    tickss << setw(done) << setfill('=') << "" << setw(rem) << setfill(' ')
           << "";

  return tickss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_tot
 * @param xfer_done
 * @return
 * @note: bytes output as measured in KiB
 */
std::string m_of_n(curl_off_t const xfer_tot, curl_off_t const xfer_done) {
  constexpr float KiB{1'024.f};
  ostringstream   m_of_nss;
  m_of_nss.imbue(locale{""});  // provides for digit separators
  m_of_nss << setw(5) << fixed << setprecision(1) << (xfer_done / KiB) << " of "
           << setw(5) << fixed << setprecision(1) << (xfer_tot / KiB);

  return m_of_nss.str();

  // https://en.cppreference.com/w/cpp/locale/locale
  // https://en.cppreference.com/w/cpp/io/basic_ios/imbue
}

//------------------------------------------------------------------------------
/**
 * @param durn
 * @return
 */
std::string hms_fmt(Durn durn) {  //

  auto const hrs = duration_cast<hours>(durn);
  durn -= hrs;
  auto const mins = duration_cast<minutes>(durn);
  durn -= mins;
  Secs_f const secs = durn;

  ostringstream hrss{};
  ostringstream minss{};

  auto const have_hrs{hrs.count() > 0};
  if (have_hrs)
    hrss << hrs.count() << 'h';

  auto const have_mins{mins.count() > 0};
  if (have_mins)
    minss << mins.count() << 'm';

  ostringstream hmss{};
  hmss << (have_hrs ? hrss.str() : "")    //
       << (have_mins ? minss.str() : "")  //
       << setw(4) << setfill(' ') << fixed << setprecision(1) << secs.count()
       << 's';

  return hmss.str();
}

//------------------------------------------------------------------------------
/**
 * @return
 * @note: simple wrapper
 */
std::string hms() {
  return hms_fmt(Call_clk::now() - start);
}

//------------------------------------------------------------------------------
/** Prints 'progress meters' for @b curl 's transfers out to console
 *
 * This function acts as a callback, called at @b curl 's discretion
 * (randomly?) while transfers are in progress
 *
 * @param unused
 * @param dl_tot
 * @param dl_rcvd
 * @param ul_tot
 * @param ul_sent
 * @return default '0', unused
 * @note: transfer's status data are supplied by the @b curl library itself,
 * we are simply providing for the required form for it here and utilizing
 * them
 */
int prt_prog(Address,
             curl_off_t dl_tot  = 0,
             curl_off_t dl_rcvd = 0,
             curl_off_t         = 0,
             curl_off_t         = 0) {  //

  if (dl_tot > 0) {  // we have a download to display

    // cast prevents integer division
    auto rcvd_frac = dl_rcvd / static_cast<float>(dl_tot);

    auto dl_pct{pct(rcvd_frac)};              // dl percent string
    auto dl_ticks{ticks(rcvd_frac, true)};    // dl tickmarks string
    auto dl_m_of_n{m_of_n(dl_tot, dl_rcvd)};  // dl bytes string

    cout << "  d/l" << dl_pct << "% [" << dl_ticks << "]  KiB : " << dl_m_of_n
         << "  " << hms() << "\r";

    // stream flushing works smoother in this particular type of case ('\r')
    cout << flush;
  }

  // finally, transfer will abort w/o a '0' func ret
  return 0;
}

}  // namespace muh_curl

// Copyright (2021)
// License (MIT) https://opensource.org/licenses/MIT
